#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<stralloc.h>
#include<string.h>

// Vergleichsfunktion fuer Strings
int stringcmp(const void *p1, const void *p2) {
	return strcmp(*(char**)p1, *(char**)p2);
}


// Hilfsfunktion zum Einlesen einer
// nullterminierten Zeile von fd nach sa
int readline(int fd, stralloc *sa) {
   if(!stralloc_copys(sa, "")) return 0;
   while(1) {
      if(!stralloc_readyplus(sa,1)) return 0;
		if(read(fd,sa->s+sa->len,1) <= 0) return 0;
      if(sa->s[sa->len] == '\n') {
			stralloc_0(sa);
			break;
		}
      sa->len++;
   }
   return 1;
}

// Hilfsfunktion zum Schreiben eines
// nullterminierten Strings s nach fd 
int writeline(int fd, char* s) {
	int written = 0;
	int len = strlen(s);
	while(written < len) {
		int bytes = write(fd, s+written, len-written);
		if(bytes == -1) {
			perror("write");
			return 0;
		}
		written += bytes;
	}
	if(write(fd, "\n", 1)<1) {
		perror("write");
		return 0;
	}
	return 1;
}


int main(int argc, char* argv[]) {
	int fdin; // Filedeskriptor der Eingabedatei
	int fdout; // Filedeskriptor der Ausgabedatei

	if(argc != 3) {
		fprintf(stderr, "Usage: %s infile outfile\n", argv[0]);
		return 1;
	}

	// Oeffne Datei 1 zum Lesen
	fdin = open(argv[1], O_RDONLY);
	if(fdin == -1) {
		perror("open");
		return 2;
	}

	// Oeffne Datei 2 zum Schreiben, ggf. ueberschreiben / neu anlegen
	fdout = open(argv[2], O_WRONLY|O_CREAT|O_TRUNC, 0644);
	if(fdout == -1) {
		perror("open");
		return 2;
	}

	int anzahl_zeilen = 0;
	int reserviert = 1;
	char** zeilen = malloc(reserviert*sizeof(char*));

	// Zeilenweise bis zum Ende der Datei lesen und Strings in den
	// Hauptspeicher sichern
	while(1) {
		stralloc sa = {0};
		if(!readline(fdin, &sa)) break;
		// Neuen Speicher fuer die Zeiger zu den einzelnen Zeilen holen
		if(anzahl_zeilen == reserviert) {
			reserviert *= 2;
			zeilen = realloc(zeilen, reserviert*sizeof(char*));
			if(zeilen == 0) {
				perror("realloc");
				return 3;
			}
		}
		// gelesene Zeile kopieren
		zeilen[anzahl_zeilen] = strdup(sa.s);
		if(zeilen[anzahl_zeilen] == 0) {
			perror("strdup");
			return 4;
		}
		anzahl_zeilen++;
		// Speicher des stralloc-Strings sa.s freigeben
		stralloc_free(&sa);
	}
	close(fdin);

	// Sortieren
	qsort(zeilen, anzahl_zeilen, sizeof(char*), stringcmp);

	// Ausgabe
	for(int i = 0; i < anzahl_zeilen; i++) {
		writeline(fdout, zeilen[i]);	
	}
	close(fdout);
	
	return 0;
}
